#!/usr/bin/env python3
"""Config-related tests."""

import os
import re

import chromaterm.config

TEMP_FILE = '.test_chromaterm_config.yml'


def test_eprint(capsys):
    """Print a message to stderr."""
    msg = 'Some random error message'
    chromaterm.config.eprint(msg)
    assert msg in capsys.readouterr().err


def test_get_color_codes():
    """Random known-good color codes."""
    colors = [
        'b#0973d8', 'b#8b6dd3', 'b#2867c7', 'b#18923a', 'b#636836', 'b#a0da5e',
        'b#b99153', 'b#fafb19', 'f#6cb0d7', 'f#6f5e3a', 'f#7d6256', 'f#15c93d',
        'f#45f2d7', 'f#50a910', 'f#1589b3', 'f#b8df23', 'f#d5a3bf', 'f#d7e764',
        'f#e002d7', 'f#f56726'
    ]
    codes = [
        '48;5;33m', '48;5;140m', '48;5;32m', '48;5;35m', '48;5;101m',
        '48;5;156m', '48;5;179m', '48;5;226m', '38;5;117m', '38;5;101m',
        '38;5;102m', '38;5;41m', '38;5;87m', '38;5;70m', '38;5;38m',
        '38;5;190m', '38;5;182m', '38;5;228m', '38;5;201m', '38;5;208m'
    ]

    for color, code in zip(colors, codes):
        color_code = chromaterm.config.get_color_codes(color)[0]['code']
        assert color_code == '\x1b[' + code


def test_get_color_codes_grayscale():
    """Random known-good grayscale codes."""
    colors = [
        'b#000000', 'b#5d5d5d', 'b#373737', 'b#c8c8c8', 'b#cecece', 'b#d7d7d7',
        'b#d8d8d8', 'b#fcfcfc', 'f#0b0b0b', 'f#000000', 'f#2b2b2b', 'f#2f2f2f',
        'f#4c4c4c', 'f#4d4d4d', 'f#9d9d9d', 'f#808080'
    ]
    codes = [
        '48;5;232m', '48;5;240m', '48;5;237m', '48;5;250m', '48;5;251m',
        '48;5;252m', '48;5;252m', '48;5;255m', '38;5;233m', '38;5;232m',
        '38;5;236m', '38;5;236m', '38;5;239m', '38;5;239m', '38;5;246m',
        '38;5;244m'
    ]

    for color, code in zip(colors, codes):
        color_code = chromaterm.config.get_color_codes(color)[0]['code']
        assert color_code == '\x1b[' + code


def test_get_color_codes_rgb():
    """RGB color-codes."""
    colors = ['b#010101', 'f#020202']
    codes = ['48;2;1;1;1m', '38;2;2;2;2m']

    for color, code in zip(colors, codes):
        color_code = chromaterm.config.get_color_codes(color, True)[0]['code']
        assert color_code == '\x1b[' + code


def test_get_color_codes_style():
    """Terminal styles."""
    colors = ['blink', 'BOLD', 'iTaLiC', 'strike', 'underline']
    codes = ['5m', '1m', '3m', '9m', '4m']

    for color, code in zip(colors, codes):
        color_code = chromaterm.config.get_color_codes(color, True)[0]['code']
        assert color_code == '\x1b[' + code


def test_get_color_codes_compound():
    """All sorts of color codes."""
    colors = 'bold b#0973d8 underline f#45f2d7'
    # Styles are always added last
    codes = ['\x1b[48;5;33m', '\x1b[38;5;87m', '\x1b[1m', '\x1b[4m']

    for color, code in zip(chromaterm.config.get_color_codes(colors), codes):
        assert color['code'] == code


def test_get_color_codes_mixed_case():
    """Color is mixed case."""
    colors = 'b#abcABC bOlD'

    assert len(chromaterm.config.get_color_codes(colors)) == 2


def test_get_color_codes_excessive_colors():
    """Too many colors (more than 2)."""
    colors = 'b#010101 f#020202 f#020202'

    assert chromaterm.config.get_color_codes(colors) is None


def test_get_color_codes_duplicate_target():
    """Duplicate targets (e.g. two foreground colors)."""
    colors = ['f#020202 f#030303', 'bold bold']

    for color in colors:
        assert chromaterm.config.get_color_codes(color) is None


def test_parse_config_simple():
    """Parse a config file with a simple rule."""
    config_data = '''rules:
    - description: simple
      regex: hello world
      color: f#fffaaa'''

    assert chromaterm.config.parse_config(config_data)['rules']


def test_parse_config_group():
    """Parse a config file with a group-specific rule."""
    config_data = '''rules:
    - regex: hello (world)! It's (me).
      color:
        1: b#fffaaa
        2: f#123123'''

    assert chromaterm.config.parse_config(config_data)['rules']


def test_parse_config_group_legacy():
    """Parse a config file with a legacy group-specific rule."""
    config_data = '''rules:
    - regex: hello (world)
      color: b#fffaaa
      group: 1'''

    assert chromaterm.config.parse_config(config_data)['rules']


def test_parse_config_multiple_colors():
    """Parse a config file with a multi-color rule."""
    config_data = '''rules:
    - description: group
      regex: hello (world)
      color: b#fffaaa f#aaafff'''

    assert chromaterm.config.parse_config(config_data)['rules']


def test_parse_config_rule_format_error(capsys):
    """Parse a config file with a syntax problem."""
    config_data = '''rules:
    - description: simple'''
    chromaterm.config.parse_config(config_data)

    assert 'Rule error on' in capsys.readouterr().err


def test_parse_config_yaml_format_error(capsys):
    """Parse an incorrect YAML file."""
    chromaterm.config.parse_config('-x-\nhi:')
    assert 'Parse error:' in capsys.readouterr().err


def test_parse_rule_regex_missing():
    """Parse a rule without a `regex` key."""
    msg = 'regex not found'
    rule = {'color': 'b#fffaaa'}
    assert chromaterm.config.parse_rule(rule) == msg


def test_parse_rule_regex_type_error():
    """Parse a rule with an incorrect `regex` value type."""
    msg = 'regex not a string'
    rule = {'regex': ['hi'], 'color': 'b#fffaaa'}
    assert chromaterm.config.parse_rule(rule) == msg

    rule = {'regex': 111, 'color': 'b#fffaaa'}
    assert chromaterm.config.parse_rule(rule) == msg


def test_parse_rule_regex_invalid():
    """Parse a rule with an invalid `regex`."""
    rule = {'regex': '+', 'color': 'b#fffaaa'}
    assert 're.error: ' in chromaterm.config.parse_rule(rule)


def test_parse_rule_color_missing():
    """Parse a rule without a `color` key."""
    msg = 'color not found'
    rule = {'regex': 'x(y)z'}
    assert chromaterm.config.parse_rule(rule) == msg


def test_parse_rule_color_type_error():
    """Parse a rule with an incorrect `color` value type."""
    msg = 'color not a string or dictionary'
    rule = {'regex': 'x(y)z', 'color': ['hi']}
    assert chromaterm.config.parse_rule(rule) == msg


def test_parse_rule_color_format_error():
    """Parse a rule with an incorrect `color` format."""
    msg_re = 'color ".+" not in the correct format'

    rule = {'regex': 'x(y)z', 'color': 'b#xyzxyz'}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))

    rule = {'regex': 'x(y)z', 'color': 'x#fffaaa'}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))

    rule = {'regex': 'x(y)z', 'color': 'b@fffaaa'}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))

    rule = {'regex': 'x(y)z', 'color': 'b#fffaaa-f#fffaaa'}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))


def test_parse_rule_group_type_error():
    """Parse a rule with an incorrect `group` value type."""
    msg_re = 'group .+ not an integer'
    rule = {'regex': 'x(y)z', 'color': {'1': 'b#fffaaa'}}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))


def test_parse_rule_group_out_of_bounds():
    """Parse a rule with `group` number not in the regex."""
    msg_re = 'group .+ not in the regex'
    rule = {'regex': 'x(y)z', 'color': {2: 'b#fffaaa'}}
    assert re.search(msg_re, chromaterm.config.parse_rule(rule))


def test_read_file():
    """Read the default configuration."""
    assert chromaterm.read_file('$HOME/.chromaterm.yml') is not None


def test_read_file_no_permission(capsys):
    """Create a file with no permissions and attempt to read it. Delete the file
    once done with it."""
    msg = 'Cannot read configuration file ' + TEMP_FILE + '1' + ' (permission)\n'

    os.close(os.open(TEMP_FILE + '1', os.O_CREAT | os.O_WRONLY, 0o0000))
    chromaterm.read_file(TEMP_FILE + '1')
    os.remove(TEMP_FILE + '1')

    assert msg in capsys.readouterr().err


def test_read_file_non_existent(capsys):
    """Read a non-existent file."""
    msg = 'Configuration file ' + TEMP_FILE + '2' + ' not found\n'
    chromaterm.read_file(TEMP_FILE + '2')
    assert msg in capsys.readouterr().err


def test_rgb_to_8bit():
    """20 random known-good translations."""
    assert chromaterm.config.rgb_to_8bit(13, 176, 1) == 40
    assert chromaterm.config.rgb_to_8bit(22, 32, 13) == 16
    assert chromaterm.config.rgb_to_8bit(29, 233, 205) == 50
    assert chromaterm.config.rgb_to_8bit(43, 48, 138) == 61
    assert chromaterm.config.rgb_to_8bit(44, 4, 5) == 52
    assert chromaterm.config.rgb_to_8bit(45, 245, 37) == 82
    assert chromaterm.config.rgb_to_8bit(75, 75, 194) == 62
    assert chromaterm.config.rgb_to_8bit(88, 121, 30) == 100
    assert chromaterm.config.rgb_to_8bit(119, 223, 223) == 123
    assert chromaterm.config.rgb_to_8bit(139, 87, 30) == 136
    assert chromaterm.config.rgb_to_8bit(146, 83, 47) == 131
    assert chromaterm.config.rgb_to_8bit(149, 67, 58) == 131
    assert chromaterm.config.rgb_to_8bit(149, 230, 209) == 158
    assert chromaterm.config.rgb_to_8bit(151, 153, 27) == 142
    assert chromaterm.config.rgb_to_8bit(163, 25, 80) == 125
    assert chromaterm.config.rgb_to_8bit(165, 186, 53) == 149
    assert chromaterm.config.rgb_to_8bit(171, 186, 6) == 184
    assert chromaterm.config.rgb_to_8bit(178, 249, 57) == 191
    assert chromaterm.config.rgb_to_8bit(229, 112, 100) == 210
    assert chromaterm.config.rgb_to_8bit(246, 240, 108) == 228



#!/usr/bin/env python3
"""Helper functions for initializing ChromaTerm's config."""

import os
import re
import sys

import yaml

# Named SGR codes
STYLES = {
    'blink': '\x1b[5m',
    'bold': '\x1b[1m',
    'italic': '\x1b[3m',
    'strike': '\x1b[9m',
    'underline': '\x1b[4m',
}

# Reset types, their default reset codes, and RegEx's for detecting color type
RESET_TYPES = {
    'fg': {
        'default': '\x1b[39m',
        're': re.compile(r'\x1b\[3(?:[0-79]|8;[0-9;]+)m')
    },
    'bg': {
        'default': '\x1b[49m',
        're': re.compile(r'\x1b\[4(?:[0-79]|8;[0-9;]+)m')
    },
    'blink': {
        'default': '\x1b[25m',
        're': re.compile(r'\x1b\[2?5m')
    },
    'bold': {
        'default': '\x1b[22m',  # Normal intensity
        're': re.compile(r'\x1b\[(?:1|2?2)m')  # Any intensity type
    },
    'italic': {
        'default': '\x1b[23m',
        're': re.compile(r'\x1b\[2?3m')
    },
    'strike': {
        'default': '\x1b[29m',
        're': re.compile(r'\x1b\[2?9m')
    },
    'underline': {
        'default': '\x1b[24m',
        're': re.compile(r'\x1b\[2?4m')
    }
}


def eprint(*args, **kwargs):
    """Error print."""
    print(sys.argv[0] + ':', *args, file=sys.stderr, **kwargs)


def get_color_codes(color_str, rgb=False):
    """Return the ANSI codes, one for each color, to be used when highlighting
    with `color_str` or None if the `color_str` is invalid."""
    color_str = color_str.lower().strip()
    color_re = r'^(((b|f)#[0-9a-f]{6}|' + '|'.join(STYLES) + r')(\s+|$))+$'

    if not re.search(color_re, color_str):
        return None

    colors = []

    # Colors
    for match in re.findall(r'(b|f)#([0-9a-f]{6})', color_str):
        if match[0] == 'f':
            target, name = '\x1b[38;', 'fg'
        else:
            target, name = '\x1b[48;', 'bg'

        if name in [x['type'] for x in colors]:  # Duplicate color target
            return None

        rgb_int = [int(match[1][i:i + 2], 16) for i in [0, 2, 4]]

        if rgb:
            target += '2;'
            color_id = ';'.join([str(x) for x in rgb_int])
        else:
            target += '5;'
            color_id = str(rgb_to_8bit(*rgb_int))

        colors.append({'code': target + color_id + 'm', 'type': name})

    # Styles
    for name in re.findall('|'.join(STYLES), color_str):
        if name in [x['type'] for x in colors]:  # Duplicate style
            return None

        colors.append({'code': STYLES[name], 'type': name})

    return colors or None


def get_default_config():
    """Return a dict with the default configuration."""
    config = {'rules': []}
    config['resets'] = {x: RESET_TYPES[x]['default'] for x in RESET_TYPES}

    return config


def parse_config(data, config=None, rgb=False):
    """Parse `data` (a YAML string), modifying/returning the `config` dict."""
    if config is None:
        config = get_default_config()

    try:  # Load the YAML configuration file
        load = yaml.safe_load(data) or {}
    except yaml.YAMLError as exception:
        eprint('Parse error:', exception)
        return config

    config['rules'] = []  # Reset the rules list
    rules = load.get('rules', []) if isinstance(load, dict) else None

    # Parse the rules
    for rule in rules if isinstance(rules, list) else []:
        parsed_rule = parse_rule(rule, rgb=rgb)
        if isinstance(parsed_rule, dict):
            config['rules'].append(parsed_rule)
        else:
            eprint('Rule error on {}: {}'.format(rule, parsed_rule))

    return config


def parse_rule(rule, rgb=False):
    """Return a dict of the description, regex (compiled), and color (code)."""
    # pylint: disable=too-many-return-statements

    description = rule.get('description', '')

    regex = rule.get('regex')
    if not regex:
        return 'regex not found'

    if not isinstance(regex, str):
        return 'regex not a string'

    try:
        regex_compiled = re.compile(regex)
    except re.error as exception:
        return 're.error: ' + str(exception)

    color = rule.get('color')
    if not color:
        return 'color not found'

    if isinstance(color, str):
        color = {0: color}
    elif not isinstance(color, dict):
        return 'color not a string or dictionary'

    for group in color:
        if not isinstance(group, int):
            return 'group "{}" not an integer'.format(group)

        if group > regex_compiled.groups:
            return 'group {} not in the regex'.format(group)

        color_codes = get_color_codes(color[group], rgb=rgb)

        if not color_codes:
            return 'color "{}" not in the correct format'.format(color)

        color[group] = color_codes

    return {
        'description': description,
        'regex': regex_compiled,
        'colors': color
    }


def read_file(location):
    """Read a file at `location`, returning its contents, or None on error."""
    location = os.path.expandvars(location)

    if not os.access(location, os.F_OK):
        eprint('Configuration file', location, 'not found')
        return None

    if not os.access(location, os.R_OK):
        eprint('Cannot read configuration file', location, '(permission)')
        return None

    with open(location, 'r') as file:
        return file.read()


def rgb_to_8bit(_r, _g, _b):
    """Downscale from 24-bit RGB to 8-bit ANSI."""
    def downscale(value, base=6):
        return int(value / 256 * base)

    if _r == _g == _b:  # Use the 24 shades of the grayscale.
        return 232 + downscale(_r, base=24)

    return 16 + (36 * downscale(_r)) + (6 * downscale(_g)) + downscale(_b)



#!/usr/bin/env python3
"""Tests for the main program."""

# You can never have too many tests, and I don't want to defy logic by moving
# code to a different file, like test_highlight.py when it lives in __init__.py.
# pylint: disable=too-many-lines

import itertools
import os
import re
import signal
import subprocess
import sys
import time
from threading import Thread

import chromaterm
import chromaterm.config

TEMP_FILE = '.test_chromaterm.yml'

CODE_ISATTY = """import os, sys
stdin = os.isatty(sys.stdin.fileno())
stdout = os.isatty(sys.stdout.fileno())
print('stdin={}, stdout={}'.format(stdin, stdout))"""

CODE_TTYNAME = """import os, sys
print(os.ttyname(sys.stdin.fileno()) if os.isatty(sys.stdin.fileno()) else None)"""

_, MOCK_FD = os.openpty()


def mock_fd_fileno():
    """Returns MOCK_FD. To be used when monkeypatching sys.stdin.fileno"""
    return MOCK_FD


def get_python_command(code):
    """Returns the python shell command that runs `code`."""
    return 'python3 -c "{}"'.format('; '.join(code.splitlines()))


def test_decode_sgr_bg():
    """Background colors and reset are being detected."""
    for code in ['\x1b[48;5;123m', '\x1b[48;2;1;1;1m', '\x1b[49m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'bg'
            assert color['code'] == code


def test_decode_sgr_fg():
    """Foreground colors and reset are being detected."""
    for code in ['\x1b[38;5;123m', '\x1b[38;2;1;1;1m', '\x1b[39m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'fg'
            assert color['code'] == code


def test_decode_sgr_styles_blink():
    """Blink and its reset are being detected."""
    for code in ['\x1b[5m', '\x1b[25m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'blink'
            assert color['code'] == code


def test_decode_sgr_styles_bold():
    """Bold and its reset are being detected."""
    for code in ['\x1b[1m', '\x1b[2m', '\x1b[22m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'bold'
            assert color['code'] == code


def test_decode_sgr_styles_italic():
    """Italic and its reset are being detected."""
    for code in ['\x1b[3m', '\x1b[23m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'italic'
            assert color['code'] == code


def test_decode_sgr_styles_strike():
    """Strike and its reset are being detected."""
    for code in ['\x1b[9m', '\x1b[29m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'strike'
            assert color['code'] == code


def test_decode_sgr_styles_underline():
    """Underline and its reset are being detected."""
    for code in ['\x1b[4m', '\x1b[24m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'underline'
            assert color['code'] == code


def test_decode_sgr_complete_reset():
    """Complete reset detection."""
    for code in ['\x1b[00m', '\x1b[0m', '\x1b[m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] == 'complete_reset'
            assert color['code'] == code


def test_decode_sgr_malformed():
    """Malformed colors."""
    for code in ['\x1b[38;5m', '\x1b[38;2;1;1m', '\x1b[38;5;123;38;2;1;1m']:
        colors = chromaterm.decode_sgr(code)
        assert len(colors) == 1

        for color in colors:
            assert color['type'] is None
            assert color['code'] == code


def test_decode_sgr_split_compound():
    """Split the a compound SGR into discrete SGR's."""
    colors = chromaterm.decode_sgr('\x1b[1;33;40m')
    codes = ['\x1b[1m', '\x1b[33m', '\x1b[40m']
    types = ['bold', 'fg', 'bg']

    for color, code, name in zip(colors, codes, types):
        assert color['type'] == name
        assert repr(color['code']) == repr(code)


def test_highlight_enscapsulated_same_type():
    """Two rules of the same target type (e.g. both foreground) and one rule
    encapsulating the other. Also tested in reverse order.
    x: --------------
    y:    ------"""
    config_data = '''rules:
    - description: first
      regex: Hello there, World
      color: f#aaafff
    - description: second
      regex: there
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;153m', 'Hello ', '\x1b[38;5;229m', 'there',
        '\x1b[38;5;153m', ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_enscapsulated_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    and one rule encapsulating the other. Also tested in reverse order.
    x: --------------
    y:    ------"""
    config_data = '''rules:
    - description: first
      regex: Hello there, World
      color: f#aaafff
    - description: second
      regex: there
      color: b#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;153m', 'Hello ', '\x1b[48;5;229m', 'there', '\x1b[49m',
        ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_partial_overlap_same_type():
    """Two rules of the same target type (e.g. both foreground) and one
    overlapping the other. Also tested in reverse order.
    x: ------
    y:   ------"""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: f#aaafff
    - description: second
      regex: there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;153m', 'Hello ', '\x1b[38;5;229m', 'there',
        '\x1b[38;5;229m', ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_partial_overlap_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    and one overlapping the other. Also tested in reverse order.
    x: ------
    y:   ------"""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: b#aaafff
    - description: second
      regex: there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[48;5;153m', 'Hello ', '\x1b[38;5;229m', 'there', '\x1b[49m',
        ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_full_overlap_same_type():
    """Two rules of the same target type (e.g. both foreground) fully overlapping
    each other. Both are applied.
    x: --------------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: Hello there, World
      color: f#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;153m', '\x1b[38;5;229m', 'Hello there, World',
        '\x1b[38;5;153m', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_full_overlap_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    fully overlapping each other. Both are applied, both ends are kept.
    x: --------------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: Hello there, World
      color: b#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[48;5;153m', '\x1b[38;5;229m', 'Hello there, World', '\x1b[39m',
        '\x1b[49m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_start_overlap_same_type():
    """Two rules of the same target type (e.g. both foreground) overlapping at
    the start. Both are applied. Also tested in reverse order. The last rule's
    color is applied closest to the match.
    x: -------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: f#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;153m', '\x1b[38;5;229m', 'Hello there', '\x1b[38;5;229m',
        ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    expected[0], expected[1] = expected[1], expected[0]  # Flip start color
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_start_overlap_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    overlapping at the start. Both are applied. Also tested in reverse order.
    The last rule's color is applied closest to the match.
    x: -------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: b#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[48;5;153m', '\x1b[38;5;229m', 'Hello there', '\x1b[49m',
        ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    expected[0], expected[1] = expected[1], expected[0]  # Flip start color
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_end_overlap_same_type():
    """Two rules of the same target type (e.g. both foreground) overlapping at
    the end. Both are applied. Also tested in reverse order. Only the first
    rule's end color is applied.
    x:        -------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: World
      color: f#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;229m', 'Hello there, ', '\x1b[38;5;153m', 'World',
        '\x1b[38;5;153m', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    expected[4] = '\x1b[38;5;229m'  # Adjust mid-color shift
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_end_overlap_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    overlapping at the end. Both are applied. Also tested in reverse order. Both
    rule's end colors are applied.
    x:        -------
    y: --------------"""
    config_data = '''rules:
    - description: first
      regex: World
      color: b#aaafff
    - description: second
      regex: Hello there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there, World'
    expected = [
        '\x1b[38;5;229m', 'Hello there, ', '\x1b[48;5;153m', 'World',
        '\x1b[39m', '\x1b[49m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    expected[-1], expected[-2] = expected[-2], expected[-1]  # Flip end color
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_end_start_overlap_same_type():
    """Two rules of the same target type (e.g. both foreground) overlapping the
    end of one with the start of the other. Both are applied. Also tested in
    reverse order.
    x: -------
    y:        -------"""
    config_data = '''rules:
    - description: first
      regex: Hello Wo
      color: f#aaafff
    - description: second
      regex: rld
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello World'
    expected = [
        '\x1b[38;5;153m', 'Hello Wo', '\x1b[39m', '\x1b[38;5;229m', 'rld',
        '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_end_start_overlap_different_type():
    """Two rules of a different target type (e.g. foreground and background)
    overlapping the end of one with the start of the other. Both are applied.
    Also tested in reverse order.
    x: -------
    y:        -------"""
    config_data = '''rules:
    - description: first
      regex: Hello Wo
      color: b#aaafff
    - description: second
      regex: rld
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello World'
    expected = [
        '\x1b[48;5;153m', 'Hello Wo', '\x1b[49m', '\x1b[38;5;229m', 'rld',
        '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))
    config['rules'] = list(reversed(config['rules']))
    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_start_same_type():
    """Highlight with an existing, same-type, color at the start of the data."""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[33mHello World'
    expected = ['\x1b[33m', '\x1b[38;5;153m', 'Hello World', '\x1b[33m']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_start_different_type():
    """Highlight with an existing, different-type, color at the start of the data."""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: b#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[33mHello World'
    expected = ['\x1b[33m', '\x1b[48;5;153m', 'Hello World', '\x1b[49m']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_end_same_type():
    """Highlight with an existing, same-type, color at the end of the data."""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello World\x1b[33m'
    expected = ['\x1b[38;5;153m', 'Hello World', '\x1b[39m', '\x1b[33m']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_end_different_type():
    """Highlight with an existing, different-type, color at the end of the data."""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: b#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello World\x1b[33m'
    expected = ['\x1b[48;5;153m', 'Hello World', '\x1b[49m', '\x1b[33m']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_multiline():
    """Highlight with an existing color in the multi-line data."""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = ['\x1b[33mHi there', 'Hello World']
    expected = [['\x1b[33m', 'Hi there'],
                ['\x1b[38;5;153m', 'Hello World', '\x1b[33m']]

    for line_data, line_expected in zip(data, expected):
        assert repr(chromaterm.highlight(config, line_data)) == repr(
            ''.join(line_expected))


def test_highlight_existing_orphaned():
    """Highlight with two existing colors in the data, the first of which was
    not closed."""
    config_data = '''rules:
    - description: first
      regex: World
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[33mHello \x1b[34mthere\x1b[m, World'
    expected = [
        '\x1b[33m', 'Hello ', '\x1b[34m', 'there', '\x1b[m', ', ',
        '\x1b[38;5;153m', 'World', '\x1b[m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_intensity_orphaned():
    """Highlight with an existing intensity (i.e. bold or faint) in the data."""
    config_data = '''rules:
    - description: first
      regex: World
      color: bold'''
    config = chromaterm.config.parse_config(config_data)

    data = ['\x1b[2mHello World', '\x1b[1mHello World']
    expected = [['\x1b[2m', 'Hello ', '\x1b[1m', 'World', '\x1b[2m'],
                ['\x1b[1m', 'Hello ', '\x1b[1m', 'World', '\x1b[1m']]

    for line_data, line_expected in zip(data, expected):
        assert repr(chromaterm.highlight(config, line_data)) == repr(
            ''.join(line_expected))


def test_highlight_existing_complete_reset():
    """Highlight with an existing complete reset in the data. It should be set
    to our color."""
    config_data = '''rules:
    - description: first
      regex: Hello there, World
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello\x1b[m there, World'
    expected = [
        '\x1b[38;5;153m', 'Hello', '\x1b[38;5;153m', ' there, World',
        '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_complete_reset_same_type():
    """Highlight with an existing color and complete reset in the data. The
    color is before the reset. The first rule's match covers both. The second
    rule's match covers the reset only. The rules are of the same type."""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: f#aaafff
    - description: first
      regex: there, World
      color: f#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[33mHello the\x1b[mre, World'
    expected = [
        '\x1b[33m', '\x1b[38;5;153m', 'Hello ', '\x1b[38;5;229m', 'the',
        '\x1b[38;5;153m', 're', '\x1b[38;5;229m', ', World', '\x1b[39m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_complete_reset_different_type():
    """Highlight with an existing color and complete reset in the data. The
    color is before the reset. The first rule's match covers both. The second
    rule's match covers the reset only. The rules are of different types."""
    config_data = '''rules:
    - description: first
      regex: Hello there
      color: f#aaafff
    - description: first
      regex: there, World
      color: b#fffaaa'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[44mHello the\x1b[mre, World'
    expected = [
        '\x1b[44m', '\x1b[38;5;153m', 'Hello ', '\x1b[48;5;229m', 'the',
        '\x1b[38;5;153m', 're', '\x1b[39m', ', World', '\x1b[44m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_existing_compound():
    """Highlight with an existing compound color (multiple SGR's) in the data."""
    config_data = '''rules:
    - regex: World|me
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = '\x1b[1;38;5;123;38;2;1;1;1;4mHello World\x1b[35;5m It\'s \x1b[32;24mme'
    expected = [
        '\x1b[1m\x1b[38;5;123m\x1b[38;2;1;1;1m\x1b[4m', 'Hello ',
        '\x1b[38;5;153m', 'World', '\x1b[38;2;1;1;1m', '\x1b[35m\x1b[5m',
        ' It\'s ', '\x1b[32m\x1b[24m', '\x1b[38;5;153m', 'me', '\x1b[32m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_partial_overlap_existing_multiline():
    """Three partially-overlapping rules with an existing color over multiple
    lines. Also tested in reverse order.
    x: ------
    y:   ------
    z:     ------"""
    config_data = '''rules:
    - description: first
      regex: Hello World
      color: f#aaafff
    - description: second
      regex: World! It's
      color: f#fffaaa
    - description: third
      regex: It's me
      color: f#0973d8'''
    config = chromaterm.config.parse_config(config_data)

    data = ['\x1b[33mSup', 'Hello World! It\'s me']
    expected = [['\x1b[33m', 'Sup'],
                [
                    '\x1b[38;5;153m', 'Hello ', '\x1b[38;5;229m', 'World',
                    '\x1b[38;5;229m', '! ', '\x1b[38;5;33m', 'It\'s',
                    '\x1b[38;5;33m', ' me', '\x1b[33m'
                ]]

    for line_data, line_expected in zip(data, expected):
        assert repr(chromaterm.highlight(config, line_data)) == repr(
            ''.join(line_expected))

    # Reset color tracking and reverse the rule order
    config = chromaterm.config.parse_config(config_data)
    config['rules'] = list(reversed(config['rules']))

    for line_data, line_expected in zip(data, expected):
        assert repr(chromaterm.highlight(config, line_data)) == repr(
            ''.join(line_expected))


def test_highlight_optional_multi_group():
    """Multiple group-specific colors."""
    config_data = '''rules:
    - regex: Hello (World)! It's (me)
      color:
        0: bold
        1: f#f7e08b
        2: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello World! It\'s me'
    expected = [
        '\x1b[1m', 'Hello ', '\x1b[38;5;229m', 'World', '\x1b[39m', '! It\'s ',
        '\x1b[38;5;153m', 'me', '\x1b[39m', '\x1b[22m'
    ]

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_optional_group_not_matched():
    """RegEx matched but the group specified is optional and did not match."""
    config_data = '''rules:
    - regex: Hello (World)?
      color:
        1: f#f7e08b'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello there! Hello World'
    expected = ['Hello there! Hello ', '\x1b[38;5;229m', 'World', '\x1b[39m']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_highlight_update_type_reset():
    """Feed a string to chromaterm that updates the config with new resets to
    confirm they are not colliding."""
    config_data = '''rules:
    - regex: Hello
      color: f#f7e08b'''
    config = chromaterm.config.parse_config(config_data)

    resets = {
        'fg': '\x1b[33m',
        'bg': '\x1b[45m',
        'blink': '\x1b[5m',
        'bold': '\x1b[1m',
        'italic': '\x1b[3m',
        'strike': '\x1b[9m',
        'underline': '\x1b[4m'
    }

    # Feed colors
    for data in resets.values():
        chromaterm.highlight(config, data)

    # Check the resets
    for name in resets:
        assert repr(config['resets'][name]) == repr(resets[name])


def test_highlight_complete_reset_defaulting_type_resets():
    """A complete SGR resets sets the type resets back to their defaults as
    defined in RESET_TYPES[x]['default']."""
    config_data = '''rules:
    - description: first
      regex: there
      color: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)
    default = chromaterm.RESET_TYPES['fg']['default']

    # Feed a color to update the type's reset
    assert chromaterm.highlight(config, '\x1b[33mHello there, \x1b[43mWorld')
    assert repr(config['resets']['fg']) == repr('\x1b[33m')

    # Feed the complete reset to send back to default
    assert chromaterm.highlight(config, '\x1b[mHello there, World')
    assert repr(config['resets']['fg']) == repr(default)


def test_highlight_zero_length_match():
    """No color should be applied to a zero-length match."""
    config_data = '''rules:
    - description: first
      regex: Hello, ()World
      color:
        1: f#aaafff'''
    config = chromaterm.config.parse_config(config_data)

    data = 'Hello, World'
    expected = ['Hello, ', 'World']

    assert repr(chromaterm.highlight(config, data)) == repr(''.join(expected))


def test_process_buffer_empty(capsys):
    """Output processing of empty input."""
    config = chromaterm.config.get_default_config()
    chromaterm.process_buffer(config, '', False)
    assert capsys.readouterr().out == ''


def test_process_buffer_more_multiple_characters(capsys):
    """Output processing when more data is hinted and input is multiple
    characters. The output should be highlighted as it isn't typing."""
    try:
        config = chromaterm.config.get_default_config()
        rule = {'regex': 'hello world', 'color': 'b#fffaaa'}
        pipe_r, pipe_w = os.pipe()

        config['read_fd'] = pipe_r
        config['rules'].append(chromaterm.config.parse_rule(rule))
        data = 'test hello world test'
        success = r'^test \x1b\[48;5;\d{1,3}mhello world\x1b\[49m test$'

        chromaterm.process_buffer(config, data, True)
        captured = capsys.readouterr()

        assert re.search(success, captured.out)
    finally:
        os.close(pipe_r)
        os.close(pipe_w)


def test_process_buffer_more_single_character(capsys):
    """Output processing when more data is hinted and input is a single
    character. The output should not highlighted as it's possibly typing."""
    try:
        config = chromaterm.config.get_default_config()
        rule = {'regex': 't', 'color': 'b#fffaaa'}
        pipe_r, pipe_w = os.pipe()

        config['read_fd'] = pipe_r
        config['rules'].append(chromaterm.config.parse_rule(rule))
        data = 't'
        success = r'^t$'

        chromaterm.process_buffer(config, data, True)
        captured = capsys.readouterr()

        assert re.search(success, captured.out)
    finally:
        os.close(pipe_r)
        os.close(pipe_w)


def test_process_buffer_multiline(capsys):
    """Output processing with multiple lines of input."""
    config = chromaterm.config.get_default_config()
    rule = {'regex': 'hello world', 'color': 'b#fffaaa'}

    config['rules'].append(chromaterm.config.parse_rule(rule))
    data = '\ntest hello world test\n'
    success = r'^test \x1b\[[34]8;5;\d{1,3}mhello world\x1b\[49m test$'

    chromaterm.process_buffer(config, data * 2, False)
    captured = capsys.readouterr()

    for line in [x for x in captured.out.splitlines() if x]:
        assert re.search(success, line)


def test_process_buffer_rule_simple(capsys):
    """Output processing with a simple rule."""
    config = chromaterm.config.get_default_config()
    rule = {'regex': 'hello world', 'color': 'b#fffaaa'}

    config['rules'].append(chromaterm.config.parse_rule(rule))
    data = 'test hello world test\n'
    success = r'^test \x1b\[48;5;\d{1,3}mhello world\x1b\[49m test$'

    chromaterm.process_buffer(config, data, False)
    captured = capsys.readouterr()

    assert re.search(success, captured.out)


def test_process_buffer_rule_group(capsys):
    """Output processing with a group-specific rule."""
    config = chromaterm.config.get_default_config()
    rule = {'regex': 'hello (world)', 'color': {1: 'b#fffaaa'}}

    config['rules'].append(chromaterm.config.parse_rule(rule))
    data = 'test hello world test\n'
    success = r'^test hello \x1b\[48;5;\d{1,3}mworld\x1b\[49m test$'

    chromaterm.process_buffer(config, data, False)
    captured = capsys.readouterr()

    assert re.search(success, captured.out)


def test_process_buffer_rule_multiple_colors(capsys):
    """Output processing with a multi-color rule."""
    config = chromaterm.config.get_default_config()
    rule = {'regex': 'hello', 'color': 'b#fffaaa f#aaafff'}

    config['rules'].append(chromaterm.config.parse_rule(rule))
    data = 'hello world\n'
    success = r'^(\x1b\[[34]8;5;\d{1,3}m){2}hello(\x1b\[[34]9m){2} world$'

    chromaterm.process_buffer(config, data, False)
    captured = capsys.readouterr()

    assert re.search(success, captured.out)


def test_read_ready_input():
    """Immediate ready when there is input buffered."""
    try:
        pipe_r, pipe_w = os.pipe()

        os.write(pipe_w, b'Hello world')
        assert chromaterm.read_ready(pipe_r)
    finally:
        os.close(pipe_r)
        os.close(pipe_w)


def test_read_ready_timeout_empty():
    """Wait with no input."""
    try:
        pipe_r, pipe_w = os.pipe()

        before = time.time()
        assert not chromaterm.read_ready(pipe_r, timeout=0.5)

        after = time.time()
        assert after - before >= 0.5
    finally:
        os.close(pipe_r)
        os.close(pipe_w)


def test_read_ready_timeout_input():
    """Immediate ready with timeout when there is input buffered."""
    try:
        pipe_r, pipe_w = os.pipe()

        os.write(pipe_w, b'Hello world')
        before = time.time()
        assert chromaterm.read_ready(pipe_r, timeout=0.5)

        after = time.time()
        assert after - before < 0.5
    finally:
        os.close(pipe_r)
        os.close(pipe_w)


def test_split_buffer_new_line_r():
    """Split based on \\r"""
    data = 'Hello \rWorld'
    expected = [['Hello ', '\r'], ['World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_new_line_r_n():
    """Split based on \\r\\n"""
    data = 'Hello \r\n World'
    expected = [['Hello ', '\r\n'], [' World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_new_line_n():
    """Split based on \\n"""
    data = 'Hello \n World'
    expected = [['Hello ', '\n'], [' World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_vertical_space():
    """Split based on \\v"""
    data = 'Hello \v World'
    expected = [['Hello ', '\v'], [' World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_form_feed():
    """Split based on \\f"""
    data = 'Hello \f World'
    expected = [['Hello ', '\f'], [' World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_c1_set():
    """Split based on the ECMA-048 C1 set, excluding CSI and OSC."""
    c1_except_csi_and_osc = itertools.chain(
        range(int('40', 16), int('5b', 16)),
        [int('5c', 16), int('5e', 16), int('5f', 16)],
    )

    for char_id in c1_except_csi_and_osc:
        data = 'Hello \x1b{} World'.format(chr(char_id))
        expected = [['Hello ', '\x1b' + chr(char_id)], [' World', '']]

        assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_exclude_csi_sgr():
    """Fail to split based on the ECMA-048 C1 CSI SGR. Added some intermediate
    characters to prevent matching other CSI codes; strictly checking empty SGR."""
    data = 'Hello \x1b[!0World'
    expected = [['Hello \x1b[!0World', '']]

    assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_csi_no_parameter_no_intermediate():
    """Split based on CSI with no parameter or intermediate bytes."""
    csi_up_to_sgr = range(int('40', 16), int('6d', 16))
    csi_above_sgr = range(int('6e', 16), int('7f', 16))

    for char_id in itertools.chain(csi_up_to_sgr, csi_above_sgr):
        data = 'Hello \x1b[{} World'.format(chr(char_id))
        expected = [['Hello ', '\x1b[' + chr(char_id)], [' World', '']]

        assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_csi_parameter_no_intermediate():
    """Split based on CSI with parameters bytes but no intermediate bytes. Up to
    3 bytes."""
    csi_up_to_sgr = range(int('40', 16), int('6d', 16))
    csi_above_sgr = range(int('6e', 16), int('7f', 16))

    for char_id in itertools.chain(csi_up_to_sgr, csi_above_sgr):
        for parameter in range(int('30', 16), int('40', 16)):
            for count in range(1, 4):
                code = chr(parameter) * count + chr(char_id)
                data = 'Hello \x1b[{} World'.format(code)
                expected = [['Hello ', '\x1b[' + code], [' World', '']]

                assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_csi_intermediate_no_parameter():
    """Split based on CSI with intermediate bytes but no parameter bytes."""
    csi_up_to_sgr = range(int('40', 16), int('6d', 16))
    csi_above_sgr = range(int('6e', 16), int('7f', 16))

    for char_id in itertools.chain(csi_up_to_sgr, csi_above_sgr):
        for intermediate in range(int('20', 16), int('30', 16)):
            for count in range(1, 4):
                code = chr(intermediate) * count + chr(char_id)
                data = 'Hello \x1b[{} World'.format(code)
                expected = [['Hello ', '\x1b[' + code], [' World', '']]

                assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_csi_parameter_intermediate():
    """Split based on CSI with parameter and intermediate bytes. Up to 3 bytes
    each."""
    csi_up_to_sgr = range(int('40', 16), int('6d', 16))
    csi_above_sgr = range(int('6e', 16), int('7f', 16))

    for char_id in itertools.chain(csi_up_to_sgr, csi_above_sgr):
        for parameter in range(int('30', 16), int('40', 16)):
            for intermediate in range(int('20', 16), int('30', 16)):
                for count in range(1, 4):
                    code = chr(parameter) * count + chr(
                        intermediate) * count + chr(char_id)
                    data = 'Hello \x1b[{} World'.format(code)
                    expected = [['Hello ', '\x1b[' + code], [' World', '']]

                    assert repr(
                        chromaterm.split_buffer(data)) == repr(expected)


def test_split_buffer_ecma_048_osc_title():
    """Operating System Command (OSC) can supply arbitrary commands within the
    visible character set."""
    for end in ['\x07', '\x1b\x5c']:
        osc = '\x1b]Ignored{}'.format(end)
        data = '{}Hello world'.format(osc)
        expected = [['', osc], ['Hello world', '']]

        assert repr(chromaterm.split_buffer(data)) == repr(expected)


def test_tty_test_code_no_pipe():
    """Baseline the test code with no pipes on stdin or stdout."""
    master, slave = os.openpty()
    subprocess.run(get_python_command(CODE_ISATTY),
                   check=True,
                   shell=True,
                   stdin=slave,
                   stdout=slave)
    assert 'stdin=True, stdout=True' in os.read(master, 1024).decode()


def test_tty_test_code_in_pipe():
    """Baseline the test code with a pipe on stdin."""
    master, slave = os.openpty()
    subprocess.run(get_python_command(CODE_ISATTY),
                   check=True,
                   shell=True,
                   stdin=subprocess.PIPE,
                   stdout=slave)
    assert 'stdin=False, stdout=True' in os.read(master, 1024).decode()


def test_tty_test_code_out_pipe():
    """Baseline the test code with a pipe on stdout."""
    _, slave = os.openpty()
    result = subprocess.run(get_python_command(CODE_ISATTY),
                            check=True,
                            shell=True,
                            stdin=slave,
                            stdout=subprocess.PIPE)
    assert 'stdin=True, stdout=False' in result.stdout.decode()


def test_tty_test_code_in_out_pipe():
    """Baseline the test code with pipes on stdin and stdout."""
    result = subprocess.run(get_python_command(CODE_ISATTY),
                            check=True,
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE)
    assert 'stdin=False, stdout=False' in result.stdout.decode()


def test_tty_test_code_ttyname_same():
    """Baseline the ttyname code, ensuring it detects matching ttys."""
    master, slave = os.openpty()

    subprocess.run(get_python_command(CODE_TTYNAME),
                   check=True,
                   shell=True,
                   stdin=slave,
                   stdout=slave)

    assert os.ttyname(slave) in os.read(master, 1024).decode()


def test_tty_test_code_ttyname_different():
    """Baseline the ttyname code, ensuring it detects different ttys."""
    master, slave = os.openpty()
    _, another_slave = os.openpty()

    subprocess.run(get_python_command(CODE_TTYNAME),
                   check=True,
                   shell=True,
                   stdin=slave,
                   stdout=slave)

    assert os.ttyname(another_slave) not in os.read(master, 1024).decode()


def test_main(capsys, monkeypatch):
    """General stdin processing."""
    try:
        pipe_r, pipe_w = os.pipe()
        monkeypatch.setattr(sys.stdin, 'fileno', mock_fd_fileno)

        config = chromaterm.args_init([])
        main_thread = Thread(target=chromaterm.main, args=(config, 1, pipe_r))

        main_thread.start()
        time.sleep(0.1)  # Any start-up delay
        assert main_thread.is_alive()

        os.write(pipe_w, b'Hello world\n')
        time.sleep(0.1)  # Any processing delay
        assert capsys.readouterr().out == 'Hello world\n'

        os.write(pipe_w, b'Hey there')
        time.sleep(0.1 + chromaterm.WAIT_FOR_SPLIT)  # Include split wait
        assert capsys.readouterr().out == 'Hey there'

        os.write(pipe_w, b'x' * (chromaterm.READ_SIZE + 1))
        time.sleep(0.1 + chromaterm.WAIT_FOR_SPLIT)  # Include split wait
        assert capsys.readouterr().out == 'x' * (chromaterm.READ_SIZE + 1)
    finally:
        os.close(pipe_r)
        os.close(pipe_w)
        main_thread.join()


def test_main_broken_pipe():
    """Break a pipe while CT is trying to write to it. The echo at the end will
    close before CT has had the chance to write to the pipe."""
    result = subprocess.run('echo | ./ct | echo',
                            check=False,
                            shell=True,
                            stderr=subprocess.PIPE)
    assert b'Broken pipe' not in result.stderr


def test_main_buffer_close_time():
    """Confirm that the program exists as soon as stdin closes."""
    before = time.time()
    subprocess.run('echo hi | ./ct', check=True, shell=True)
    after = time.time()

    assert after - before < 1


def test_main_decode_error(capsys, monkeypatch):
    """Attempt to decode a character that is not UTF-8."""
    try:
        pipe_r, pipe_w = os.pipe()
        monkeypatch.setattr(sys.stdin, 'fileno', mock_fd_fileno)

        config = chromaterm.args_init([])
        main_thread = Thread(target=chromaterm.main, args=(config, 1, pipe_r))

        main_thread.start()
        time.sleep(0.1)  # Any start-up delay
        assert main_thread.is_alive()

        os.write(pipe_w, b'\x80')
        time.sleep(0.1)  # Any processing delay
        assert capsys.readouterr().out == '�'
    finally:
        os.close(pipe_r)
        os.close(pipe_w)
        main_thread.join()


def test_main_reload_config(capsys, monkeypatch):
    """Reload the configuration while the program is running."""
    try:
        with open(TEMP_FILE + '1', 'w') as file:
            file.write('''rules:
            - regex: Hello
              color: f#123123
            - regex: world
              color: b#321321''')

        pipe_r, pipe_w = os.pipe()
        monkeypatch.setattr(sys.stdin, 'fileno', mock_fd_fileno)

        # For deterministic color codes, disable RGB terminal if supported
        os.environ.pop('COLORTERM', None)
        config = chromaterm.args_init(['--config', TEMP_FILE + '1'])
        main_thread = Thread(target=chromaterm.main, args=(config, 1, pipe_r))

        main_thread.start()
        time.sleep(0.1)  # Any start-up delay
        assert main_thread.is_alive()

        os.write(pipe_w, b'Hello world\n')
        expected = [
            '\x1b[38;5;22m', 'Hello', '\x1b[39m', ' ', '\x1b[48;5;52m',
            'world', '\x1b[49m', '\n'
        ]
        time.sleep(0.1)  # Any processing delay
        assert repr(capsys.readouterr().out) == repr(''.join(expected))

        # Create file without the 'world' rule
        os.remove(TEMP_FILE + '1')
        with open(TEMP_FILE + '1', 'w') as file:
            file.write('''rules:
            - regex: Hello
              color: f#123123''')

        # Reload config
        os.kill(os.getpid(), signal.SIGUSR1)

        os.write(pipe_w, b'Hello world\n')
        expected = ['\x1b[38;5;22m', 'Hello', '\x1b[39m', ' world\n']
        time.sleep(0.1)  # Any processing delay
        assert repr(capsys.readouterr().out) == repr(''.join(expected))
    finally:
        os.close(pipe_r)
        os.close(pipe_w)
        os.remove(TEMP_FILE + '1')
        main_thread.join()


def test_main_reload_processes():
    """Reload all other CT processes."""
    processes = [  # Spawn dummy ct processes
        subprocess.Popen('sleep 1 | ./ct', shell=True) for _ in range(3)
    ]

    program = ['./ct', '--reload']
    result = subprocess.run(program, check=False, stderr=subprocess.PIPE)
    assert result.stderr == b'Processes reloaded: 3\n'

    for process in processes:
        process.wait()


def test_main_run_no_file_found():
    """Have CT run with an unavailable command."""
    program = ['./ct', 'plz-no-work']
    result = subprocess.run(program, check=False, stdout=subprocess.PIPE)
    output = re.sub(br'\x1b\[[\d;]+?m', b'', result.stdout)
    assert b'./ct: plz-no-work: command not found' in output


def test_main_run_no_pipe():
    """Have CT run the tty test code with no pipes."""
    master, slave = os.openpty()
    subprocess.run('./ct ' + get_python_command(CODE_ISATTY),
                   check=True,
                   shell=True,
                   stdin=slave,
                   stdout=slave)
    assert 'stdin=True, stdout=True' in os.read(master, 1024).decode()


def test_main_run_in_pipe():
    """Have CT run the tty test code with a pipe on stdin."""
    master, slave = os.openpty()
    subprocess.run('./ct ' + get_python_command(CODE_ISATTY),
                   check=True,
                   shell=True,
                   stdin=subprocess.PIPE,
                   stdout=slave)
    assert 'stdin=True, stdout=True' in os.read(master, 1024).decode()


def test_main_run_out_pipe():
    """Have CT run the tty test code with a pipe on stdout."""
    _, slave = os.openpty()
    result = subprocess.run('./ct ' + get_python_command(CODE_ISATTY),
                            check=True,
                            shell=True,
                            stdin=slave,
                            stdout=subprocess.PIPE)
    assert 'stdin=True, stdout=True' in result.stdout.decode()


def test_main_run_in_out_pipe():
    """Have CT run the tty test code with pipes on stdin and stdout."""
    result = subprocess.run('./ct ' + get_python_command(CODE_ISATTY),
                            check=True,
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE)
    assert 'stdin=True, stdout=True' in result.stdout.decode()


def test_main_run_child_ttyname():
    """Ensure that CT spawns the child in a pseudo-terminal."""
    master, slave = os.openpty()

    subprocess.run('./ct ' + get_python_command(CODE_TTYNAME),
                   check=True,
                   shell=True,
                   stdin=slave,
                   stdout=slave)

    assert os.ttyname(slave) not in os.read(master, 1024).decode()



#!/usr/bin/env python3
"""Miscellaneous functions unrelated to primary functions of ChromaTerm."""

import os

DEFAULT_CONFIG = r"""rules:
- description: IPv4
  regex: \b(?<!\.)((25[0-5]|(2[0-4]|[0-1]?\d)?\d)\.){3}(25[0-5]|(2[0-4]|[0-1]?\d)?\d)(/\d+)?(?!\.)\b
  color: f#00ffff

- description: IPv6 (boundaries don't work here as they can be in the start or end of the match, so using lookaheads and lookbehinds instead)
  regex: (?i)((?<=[\W])|^)(?<!:)(([\da-f]{1,4}:){7}[\da-f]{1,4}|([\da-f]{1,4}:){1,1}(:[\da-f]{1,4}){1,6}|([\da-f]{1,4}:){1,2}(:[\da-f]{1,4}){1,5}|([\da-f]{1,4}:){1,3}(:[\da-f]{1,4}){1,4}|([\da-f]{1,4}:){1,4}(:[\da-f]{1,4}){1,3}|([\da-f]{1,4}:){1,5}(:[\da-f]{1,4}){1,2}|([\da-f]{1,4}:){1,6}(:[\da-f]{1,4})|([\da-f]{1,4}:){1,7}:|:((:[\da-f]{1,4}){1,7}|:)|::(ffff(:0{1,4})?:)?((25[0-5]|(2[0-4]|[0-1]?\d)?\d)\.){3}(25[0-5]|(2[0-4]|[0-1]?\d)?\d)|([\da-f]{1,4}:){1,4}:((25[0-5]|(2[0-4]|[0-1]?\d)?\d)\.){3}(25[0-5]|(2[0-4]|[0-1]?\d)?\d))(%[\da-z]+)?(/\d+)?(?!:)((?=[\W])|$)
  color: f#ff00ff

- description: MAC addresses
  regex: (?i)\b((?<!:)([\da-f]{1,2}:){5}[\da-f]{1,2}(?!:)|(?<!\.)([\da-f]{4}\.){2}[\da-f]{4}(?!\.))\b
  color: f#5f61ad

- description: Date in YYYY-MM-DD, YY-MM-DD, MMM (YYYY|DD), or DD MMM (YYYY)? formats
  regex: (?i)((?<=[\W])|^)((\d{2}|\d{4})\-((0)?[1-9]|1[0-2])\-(3[0-1]|([1-2]\d)|(0)?[1-9])|(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s(\d{4}|\s\d|(3[0-1]|([0-2]\d)))|((\d|(3[0-1]|([0-2]\d)))\s(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)(\s\d{4})?))((?=[\WT_])|$)
  color: b#af5f00

- description: Time in hh:mm:ss.SSSSSS format (sec, msec, and nsec optional)
  regex: \b((?<!:)((2[0-3])|[0-1]\d):[0-5]\d(:[0-5]\d)?((\.|,)\d{3,6})?(?!:))\b
  color: b#af5f00

- description: Generics - Bad
  regex: (?i)\b(password|abnormal(ly)?|down|los(t|s|ing)|err(or(s)?)?|(den(y|ies|ied)?)|reject(ing|ed)?|drop(ped|s)?|(err\-)?disable(d)?|(time(d)?(\-)?out)|fail(s|ed|iure)?|disconnect(ed)?|unreachable|invalid|bad|notconnect|unusable|block(ing|ed)?|blk|inaccessible|wrong|collision(s)?|unsynchronized|mismatch|runts|CRC)\b
  color: f#ff0000

- description: Generics - Ambigious bad
  regex: (?i)\b(no(t|pe)?|exit(ed)?|reset((t)?ing)?|discard(ed|ing)?|filter(ed)?|stop(p(ed|ing))?|never|can((')?t|not))\b
  color: f#865e12

- description: Generics - Not too bad
  regex: (?i)\b(warning(s)?)\b
  color: f#ffff00

- description: Generics - Ambigious good
  regex: (?i)\b(ye(s|a(h)?|p)?|started|running|can)\b
  color: f#085e0b

- description: Generics - Good
  regex: (?i)\b(up|ok(ay)?|permit(ed|s)?|accept(s|ed)?|enable(d)?|online|succe((ss(ful|fully)?)|ed(ed)?)?|connect(ed)?|reachable|valid|forwarding|synchronized)\b
  color: f#00ff00
"""


def write_default_config(directory=os.getenv('HOME'), name='.chromaterm.yml'):
    """Write the default configuration file if it doesn't exist."""
    if not directory or not name:
        return False

    location = os.path.join(directory, name)

    if os.access(location, os.F_OK):  # Already exists
        return False

    if not os.access(directory, os.W_OK):  # No write permission in directory
        return False

    with open(location, 'w') as file:
        file.write(DEFAULT_CONFIG)

    return True



#!/usr/bin/env python3
"""Colorize your output using RegEx."""

# A couple of sections of the program are used _rarely_ and I don't want to
# _always_ spend time importing them.
# pylint: disable=import-outside-toplevel

import argparse
import os
import re
import select
import signal
import sys

from .config import RESET_TYPES, eprint, parse_config, read_file

# Sequences upon which ct will split during processing. This includes new lines,
# vertical spaces, form feeds, C1 set (ECMA-048), CSI (excluding SGR), and OSC.
SPLIT_RE = re.compile(r'(\r\n?|\n|\v|\f|\x1b[\x40-\x5a\x5c\x5e\x5f]|'
                      r'\x1b\x5b[\x30-\x3f]*[\x20-\x2f]*[\x40-\x6c\x6e-\x7e]|'
                      r'\x1b\x5d[\x08-\x0d\x20-\x7e]*(?:\x07|\x1b\x5c))')

# Select Graphic Rendition sequence (all types)
SGR_RE = re.compile(r'\x1b\[[0-9;]*m')

# Maximum chuck size per read
READ_SIZE = 4096  # 4 KiB

# CT cannot determine if it is processing input faster than the piping process
# is outputting or if the input has finished. To work around this, CT will wait
# a bit prior to assuming there's no more data in the buffer. There's no impact
# on performance as the wait is cancelled if read_fd becomes ready. 1/256 is
# smaller (shorter) than the fastest key repeat (1/255 second).
WAIT_FOR_SPLIT = 1 / 256


def args_init(args=None):
    """Return the parsed configuration according to the program arguments. if
    there is an error, a string with the message is returned."""
    epilog = """ChromaTerm reads from standard input; just pipe data to it. As
    this exposes the existance of a pipe to the piping process, ChromaTerm
    can run your program in order to hide the pipe. This is normally only
    needed for programs that want to be on a controlling terminal, like `less`."""

    parser = argparse.ArgumentParser(epilog=epilog)

    parser.add_argument('program',
                        type=str,
                        metavar='program ...',
                        nargs='?',
                        help='run a program with anything after it used as '
                        'arguments')
    parser.add_argument('arguments',
                        type=str,
                        nargs=argparse.REMAINDER,
                        help=argparse.SUPPRESS,
                        default=[])
    parser.add_argument('--config',
                        metavar='FILE',
                        type=str,
                        help='location of config file (default: %(default)s)',
                        default='$HOME/.chromaterm.yml')
    parser.add_argument('--reload',
                        action='store_true',
                        help='Reload the config of all CT instances')
    parser.add_argument('--rgb',
                        action='store_true',
                        help='Use RGB colors (default: detect support, '
                        'fallback to xterm-256)')

    args = parser.parse_args(args)

    if args.reload:
        import psutil

        count = 0
        current_process = psutil.Process()

        for process in psutil.process_iter():
            if process.pid == current_process.pid:  # Skip the current process
                continue

            try:
                # Only compare the first two arguments (Python and script paths)
                if process.cmdline()[:2] == current_process.cmdline()[:2]:
                    os.kill(process.pid, signal.SIGUSR1)
                    count += 1
            except (psutil.AccessDenied, psutil.NoSuchProcess):
                # As per the documentation, expect those errors when accessing
                # the methods of process
                pass

        sys.exit('Processes reloaded: ' + str(count))

    rgb = os.getenv('COLORTERM') == 'truecolor' or args.rgb
    config = parse_config(read_file(args.config) or '', rgb=rgb)

    def update_config_handler(_1, _2):
        parse_config(read_file(args.config) or '', config, rgb)

    if args.program:
        config['read_fd'] = run_program([args.program] + args.arguments)

    signal.signal(signal.SIGPIPE, signal.SIG_DFL)  # Default for broken pipe
    signal.signal(signal.SIGINT, signal.SIG_IGN)  # Ignore SIGINT
    signal.signal(signal.SIGUSR1, update_config_handler)  # Reload handler

    return config


def decode_sgr(source_code):
    """Decode an SGR, splitting it into discrete colors."""
    def make_sgr(code_id):
        return '\x1b[' + str(code_id) + 'm'

    colors = []
    codes = source_code.lstrip('\x1b[').rstrip('m').split(';')
    skip = 0

    for index, code in enumerate(codes):
        if skip:  # Code processed by an index look-ahead; skip it
            skip -= 1
            continue

        if code == '' or int(code) == 0:  # Complete reset
            colors.append({'code': make_sgr(code), 'type': 'complete_reset'})
        elif code in ['38', '48']:  # Multi-code SGR
            color_type = 'fg' if code == '38' else 'bg'

            if len(codes) > index + 2 and codes[index + 1] == '5':  # xterm-256
                skip = 2
                code = ';'.join([str(codes[index + x]) for x in range(3)])
            elif len(codes) > index + 4 and codes[index + 1] == '2':  # RGB
                skip = 4
                code = ';'.join([str(codes[index + x]) for x in range(5)])
            else:  # Does not conform to format; do not touch code
                return [{'code': source_code, 'type': None}]

            colors.append({'code': make_sgr(code), 'type': color_type})
        else:  # Single-code SGR
            for name in RESET_TYPES:
                if RESET_TYPES[name]['re'].search(make_sgr(int(code))):
                    colors.append({'code': make_sgr(code), 'type': name})

    return colors


def get_rule_inserts(rule, data):
    """Return a list of dicts, with each dict containing the start position, end
    position, and color of a match."""
    inserts = []

    for match in rule['regex'].finditer(data):
        for group in rule['colors']:
            # Group is zero-length or is not in the match (start = end = -1)
            if match.start(group) == match.end(group):
                continue

            inserts.append({
                'start': match.start(group),
                'end': match.end(group),
                'colors': rule['colors'][group]
            })

    return inserts


def highlight(config, data):
    """According to the rules in the `config`, return the highlighted 'data'.
    The resets of each color type in the config are updated accordingly."""
    if not data:  # Empty data, don't bother doing anything
        return data

    existing, data = strip_colors(data)

    inserts = []  # The list of colors and their positions (inserts)

    for rule in config['rules']:  # Get the list of the new colors
        inserts += get_rule_inserts(rule, data)

    # Process all of the inserts, returning the final list, including existing
    inserts = process_inserts(inserts, existing, config)
    updated_resets = []

    for insert in inserts:  # Insert the colors into the data
        index = insert['position']
        data = data[:index] + insert['code'] + data[index:]

        # Update the resets according to the last reset of each type
        for name in [x for x in config['resets'] if x not in updated_resets]:
            if insert['type'] == 'complete_reset':
                # Set to type's default reset on a complete reset
                config['resets'][name] = RESET_TYPES[name]['default']
                updated_resets.append(name)
            elif name == insert['type']:
                config['resets'][name] = insert['code']
                updated_resets.append(name)

    return data


def process_buffer(config, buffer, more):
    """Process the `buffer` using the `config`, returning any left-over data. If
    there's `more`, only process up to the last split. Otherwise, process all of
    the buffer."""
    if not buffer or not config['rules']:  # Nothing to do
        print(buffer, end='')
        return ''

    splits = split_buffer(buffer)

    for split in splits[:-1]:  # Process all splits except for the last
        print(highlight(config, split[0]) + split[1], end='')

    # Indicated more data to possibly come and read_fd confirmed it
    if more and read_ready(config.get('read_fd'), timeout=WAIT_FOR_SPLIT):
        # Return last split as the left-over data
        return splits[-1][0] + splits[-1][1]

    # A single character would indicate keyboard typing; don't highlight. This
    # works with WAIT_FOR_SPLIT which will split keyboard typing accordingly.
    if len(splits[-1][0]) == 1:
        leftover_data = splits[-1][0]
    else:
        leftover_data = highlight(config, splits[-1][0])

    # No more data; print last split and flush (doesn't have a new line).
    print(leftover_data + splits[-1][1], end='', flush=True)

    return ''  # All of the buffer was processed; return an empty buffer


def process_inserts(inserts, existing, config):
    """Process a list of rule inserts, removing any unnecessary colors, and
    returning an iterator of colors. The list of existing colors is used for
    recovery of colliding colors. Existing colors are included in the iterator."""
    def get_last_color(colors, position, color_type):
        """Return the first color before the requested position and of the color
        type, or None."""
        color_types = [color_type, 'complete_reset']

        for color in reversed(sorted(colors, key=lambda x: x['position'])):
            if color['position'] < position and color['type'] in color_types:
                return color
        return None

    finals = existing

    for insert in inserts:
        for color in insert['colors']:
            # Get the last color prior to adding current color to the final list
            last_color = get_last_color(finals, insert['end'], color['type'])

            finals.append({
                'position': insert['start'],
                'code': color['code'],
                'type': color['type']
            })

            if not last_color:  # No last color; use current type reset
                reset = config['resets'][color['type']]
                color_type = color['type']
            else:
                reset = last_color['code']

                # Last color is in the middle of current color.
                if last_color['position'] > insert['start']:
                    # Last color is a (complete )?reset
                    if last_color['type'] == 'complete_reset' or last_color[
                            'code'] == config['resets'][color['type']]:
                        # Set it to our color so that it doesn't reset it.
                        last_color['code'] = color['code']
                        last_color['type'] = color['type']
                        reset = config['resets'][color['type']]

                color_type = last_color['type']

            # Post-reverse, end of the current color comes after previous ends
            finals.insert(0, {
                'position': insert['end'],
                'code': reset,
                'type': color_type
            })

    # Sort from end to start (index magic). Must reverse after sorting, not during
    return reversed(sorted(finals, key=lambda x: x['position']))


def read_ready(*read_fds, timeout=None):
    """Wrapper for the read fds of `select`. Returns the list of ready read fds."""
    return [] if not read_fds else select.select(read_fds, [], [], timeout)[0]


def run_program(program_args):
    """Fork a program over a pty of which the master fd is returned."""
    import atexit
    import fcntl
    import termios
    import pty
    import shutil
    import struct
    import tty

    # Save the current tty's window size and attributes (used by slave pty)
    window_size = shutil.get_terminal_size()
    window_size = struct.pack('2H', window_size.lines, window_size.columns)

    try:
        attributes = termios.tcgetattr(sys.stdin.fileno())

        # Set to raw as the pty will be handling any processing
        tty.setraw(sys.stdin.fileno())
        atexit.register(termios.tcsetattr, sys.stdin.fileno(), termios.TCSANOW,
                        attributes)
    except termios.error:
        attributes = None

    pid, master_fd = pty.fork()  # openpty, login_tty, then fork

    if pid == 0:  # Program
        # Update the slave's pty (now on std fds) window size and attributes
        fcntl.ioctl(sys.stdin.fileno(), termios.TIOCSWINSZ, window_size)

        if attributes:
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, attributes)

        try:
            os.execvp(program_args[0], program_args)
        except FileNotFoundError:
            eprint(program_args[0] + ': command not found')

        sys.exit(1)  # Shouldn't be hit as exec replaces the fork's process
    else:  # CT
        return master_fd


def split_buffer(buffer):
    """Split the buffer based on split sequences, returning a list of lists in
    the format of [data, separator]. `data` should be highlighted while the
    `sperator` is printed unchanged, after the `data`."""
    splits = SPLIT_RE.split(buffer)

    # Append an empty separator in case of no splits or no separator at the end
    splits.append('')

    # Group all splits into format of [data, separator]
    return [[x, y] for x, y in zip(splits[0::2], splits[1::2])]


def strip_colors(data):
    """Remove the colors from the data, returning a list of colors as well as
    the clean data. The color positions are relative to the "clean" data."""
    colors = []

    while True:
        match = SGR_RE.search(data)  # Get the first match

        if not match:  # Stop if there aren't any SGR's in the data
            break

        for color in decode_sgr(match.group()):  # Split compound colors
            color['position'] = match.start()
            colors.append(color)

        # Remove match from data; next match's start is in the clean data
        data = data[:match.start()] + data[match.end():]

    return colors, data


def main(config, max_wait=None, read_fd=None):
    """Main entry point that uses `config` from args_init to process data.
    `max_wait` is the longest period to wait without input before returning.
    read_fd will default to stdin if not specified. If config['read_fd'] is set,
    the read_fd keyword is ignored."""
    # data piped over stdin
    if config.get('read_fd', read_fd) in [None, sys.stdin.fileno()]:
        config['read_fd'] = sys.stdin.fileno()
        fds = [config['read_fd']]
    # data coming from pty, stdin is forwarded to the pty
    else:
        config['read_fd'] = config.get('read_fd', read_fd)
        fds = [sys.stdin.fileno(), config['read_fd']]

    buffer = ''
    ready_fds = read_ready(*fds, timeout=max_wait)

    while ready_fds:
        # stdin has data (piped) or has input to forward to the pty
        if sys.stdin.fileno() in ready_fds:
            try:
                data = os.read(sys.stdin.fileno(), READ_SIZE)

                if sys.stdin.fileno() == config['read_fd']:  # stdin is data
                    buffer += data.decode(encoding='utf-8', errors='replace')
                else:  # stdin is forwarded to pty
                    os.write(config['read_fd'], data)
            except OSError:
                data = b''

            if not data:  # stdin or pty closed; don't forward anymore
                fds.remove(sys.stdin.fileno())

        # Data was received. If read_fd is stdin, then just process the buffer
        # as the data was already read above. Otherwise, read_fd is pty.
        if config['read_fd'] in ready_fds:
            if config['read_fd'] != sys.stdin.fileno():
                try:
                    data = os.read(config['read_fd'], READ_SIZE)
                except OSError:
                    data = b''

                buffer += data.decode(encoding='utf-8', errors='replace')

            if not buffer:  # Buffer was processed empty and data fd hit EOF
                break

            # Process the buffer, updating it with any left-over data
            buffer = process_buffer(config, buffer, bool(data))

        ready_fds = read_ready(*fds, timeout=max_wait)
